package org.dasein.persist;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;
import org.dasein.util.CachedItem;
import org.dasein.util.ConcurrentMultiCache;
import org.dasein.util.JiteratorFilter;

public abstract class PersistentCache<T extends CachedItem> {
    static private final Logger logger = Logger.getLogger(PersistentCache.class);
    
    static private final HashMap<String,PersistentCache<? extends CachedItem>> caches = new HashMap<String,PersistentCache<? extends CachedItem>>();
    
    @SuppressWarnings("unchecked")
    static public PersistentCache<? extends CachedItem> getCache(Class<? extends CachedItem> forClass, String primaryKey) throws PersistenceException {
        PersistentCache<? extends CachedItem> cache = null;
        String className = forClass.getName();
        
        synchronized( caches ) {
            cache = caches.get(className);
            if( cache != null ) {
                return cache;
            }
        }
        
        Properties props = new Properties();
        
        try {
            InputStream is = DaseinSequencer.class.getResourceAsStream(DaseinSequencer.PROPERTIES);

            if( is != null ) {
                props.load(is);
            }
        }
        catch( Exception e ) {
            e.printStackTrace();
        }
        String key = "dsn.secondaryKeys." + className;
        String prop = props.getProperty(key);
        Key[] withKeys;
        
        if( prop == null ) {
            withKeys = new Key[] { new Key(primaryKey) };
        }
        else {
            String[] keyList = prop.split(";");
            
            if( keyList == null || keyList.length < 2 ) {
                withKeys = new Key[] { new Key(primaryKey), new Key(prop) };
            }
            else {
                int idx = 1;
                
                withKeys = new Key[keyList.length + 1];
                withKeys[0] = new Key(primaryKey);
                for( String k : keyList ) {
                    String[] parts = k.split(",");
                    
                    if( parts == null || parts.length < 2 ) {
                        withKeys[idx++] = new Key(k);
                    }
                    else {
                        withKeys[idx++] = new Key(parts);
                    }
                }
            }
        }
        key = "dsn.persistentCache." + className;
        while( cache != null && !key.equals("dsn.persistentCache") ) {
            prop = props.getProperty(key);
            if( prop != null ) {
                try {
                    Class<? extends PersistentCache<?>> cls = (Class<? extends PersistentCache<?>>)Class.forName(prop);
                    Constructor<? extends PersistentCache<?>> constructor = cls.getConstructor(forClass, Key[].class);
                    
                    cache = constructor.newInstance(forClass, withKeys);
                    break;
                }
                catch( Throwable t ) {
                    logger.error("Unable to load persistence cache " + prop + ": " + t.getMessage());
                    throw new PersistenceException("Unable to load persistence cache " + prop + ": " + t.getMessage());
                }
            }
            int idx = key.lastIndexOf('.');
                
            key = key.substring(0,idx);
        }
        if( cache == null ) {
            prop = props.getProperty("dsn.cache.default");
            if( prop == null ) {
                throw new PersistenceException("No persistent cache implementations defined.");
            }
            try {
                Class<? extends PersistentCache<?>> cls = (Class<? extends PersistentCache<?>>)Class.forName(prop);
                Constructor<? extends PersistentCache<?>> constructor = cls.getConstructor(forClass, Key[].class);
                
                cache = constructor.newInstance(forClass, withKeys);
            }
            catch( Throwable t ) {
                logger.error("Unable to load persistence cache " + prop + ": " + t.getMessage());
                throw new PersistenceException("Unable to load persistence cache " + prop + ": " + t.getMessage());
            }            
        }
        synchronized( caches ) {
            PersistentCache<? extends CachedItem> c = caches.get(className);
            
            if( c != null ) {
                cache = c;
            }
            else {
                caches.put(className, cache);
            }
        }
        return cache;
    }
    
    private ConcurrentMultiCache<T> cache         = null;
    private Key                     primaryKey    = null;
    private Key[]                   secondaryKeys = null;
    
    protected PersistentCache(Class<T> cls, Key ... keys) {
        primaryKey = keys[0];
        secondaryKeys = new Key[keys.length-1];
        if( keys.length > 1 ) {
            for( int i=1; i< keys.length; i++ ) {
                secondaryKeys[i-1] = keys[i];
            }
        }  
        if( secondaryKeys.length > 0 ) {
            ArrayList<String> solos = new ArrayList<String>();
        
            solos.add(primaryKey.getFields()[0]);
            for( Key key : secondaryKeys ) {
                if( key.getFields().length == 1 ) {
                    solos.add(key.getFields()[0]);
                }
            }
            cache = new ConcurrentMultiCache<T>(cls, solos.toArray(new String[solos.size()]));
        }
        else {
            cache = new ConcurrentMultiCache<T>(cls, primaryKey.getFields()[0]);
        }
    }
    
    public long count() throws PersistenceException {
        return list().size();
    }
    
    public long count(SearchTerm ... terms) throws PersistenceException {
        return find(terms).size();
    }
    
    protected ConcurrentMultiCache<T> getCache() {
        return cache;
    }
    
    public abstract T create(Transaction xaction, Map<String,Object> state) throws PersistenceException;
    
    public Collection<T> find(SearchTerm ... terms) throws PersistenceException {
        return find(terms, null, false);
    }
    
    public abstract Collection<T> find(SearchTerm[] terms, JiteratorFilter<T> filter, Boolean orderDesc, String ... orderFields) throws PersistenceException;
    
    public abstract T get(Object keyValue) throws PersistenceException;
    
    protected String getKeyValue(T object) throws PersistenceException {
        return getKeyValue(object, getPrimaryKey());
    }
    
    @SuppressWarnings("unchecked")
    protected String getKeyValue(T object, Key key) throws PersistenceException {
        if( object == null ) {
            return null;
        }
        try {
            String value = "";
            
            for( String fieldName : key.getFields() ) {
                Class<? extends Object> t = getCache().getTarget();
                Field field = null;
                
                while( field == null ) {
                    field = t.getDeclaredField(fieldName);
                    if( field == null ) {
                        t = t.getSuperclass();
                        if( t.getName().equals(Object.class.getName()) ) {
                            return null;
                        }
                    }
                }
                field.setAccessible(true);
                if( !value.equals("") ) {
                    value = value + ":";
                }
                Object result = field.get(object);
                
                if( result instanceof Enum ) {
                    value = value + ((Enum)result).name();
                }
                else {
                    value = value + result.toString();
                }
            }
            return value;
        }
        catch( Exception e ) {
            e.printStackTrace();
            throw new PersistenceException(e);
        }
    }
    
    @SuppressWarnings("unchecked")
    protected String getKeyValue(Map<String,Object> map, Key key) throws PersistenceException {
        if( map == null ) {
            return null;
        }
        try {
            String value = "";
            
            for( String fieldName : key.getFields() ) {
                if( !value.equals("") ) {
                    value = value + ":";
                }
                Object result = map.get(fieldName);
                
                if( result instanceof Enum ) {
                    value = value + ((Enum)result).name();
                }
                else {
                    value = value + result.toString();
                }
            }
            return value;
        }
        catch( Exception e ) {
            e.printStackTrace();
            throw new PersistenceException(e);
        }
    }
    
    @SuppressWarnings("unchecked")
    protected Set<String> getKeyValues(int index, SearchTerm ... forTerms) {
        HashSet<String> list = new HashSet<String>();
        SearchTerm term = forTerms[index];
        Object searchValue = term.getValue();
        String keyValue;
        
        if( searchValue instanceof Range ) {
            keyValue = null;
        }
        else if( searchValue instanceof Enum ) {
            keyValue = ((Enum<?>)searchValue).name();
        }
        else {
            keyValue = searchValue.toString();
        }
        if( index < forTerms.length-1 ) {
            for( String remainder : getKeyValues(index+1, forTerms) ) {
                if( searchValue instanceof Range ) {
                    for( int val : (Range)searchValue ) {
                        list.add(String.valueOf(val) + ":" + remainder);
                    }
                }
                else {
                    list.add(keyValue + ":" + remainder);
                }
            }
        }
        else {
            if( searchValue instanceof Range ) {
                for( int val : (Range)searchValue ) {
                    list.add(String.valueOf(val));
                }
            }
            else {
                list.add(keyValue);
            }
        }
        return list;
    }
    
    protected Key getPrimaryKey() {
        return primaryKey;
    }
    
    protected String getPrimaryKeyField() {
        return primaryKey.getFields()[0];
    }
    
    public long getNewKeyValue() throws PersistenceException {
        return Sequencer.getInstance(getTarget().getName() + "." + getPrimaryKeyField()).next();
    }
    
    protected Key[] getSecondaryKeys() {
        return secondaryKeys;
    }
    
    public Class<T> getTarget() {
        return cache.getTarget();
    }
    
    public abstract Collection<T> list() throws PersistenceException;
    
    public Collection<T> list(boolean orderDesc, String ... orderFields) throws PersistenceException {
        if( !orderDesc || (orderFields != null && orderFields.length > 0 ) ) {
            return find(new SearchTerm[0], null, orderDesc, orderFields);
        }
        else {
            return list();
        }
    }
    
    protected Key matchKeys(SearchTerm[] terms) {
        if( terms == null ) {
            return null;
        }
        if( terms.length == 1 && terms[0].getColumn().equals(primaryKey.getFields()[0]) ) {
            return primaryKey ;
        }
        for( Key key : secondaryKeys ) {
            if( key.matches(terms) ) {
                return key;
            }
        }
        return null;
    }
    
    public abstract void remove(Transaction xaction, T item) throws PersistenceException;
    
    public abstract void update(Transaction xaction, T item, Map<String,Object> state) throws PersistenceException;        
}
