package org.dasein.persist;

import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.UUID;

import net.sf.json.JSONObject;

import org.apache.log4j.Logger;

import org.dasein.util.CachedItem;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorFilter;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;
import org.json.JSONArray;
import org.json.JSONException;

import com.basho.riak.client.RiakClient;
import com.basho.riak.client.RiakLink;
import com.basho.riak.client.RiakObject;
import com.basho.riak.client.mapreduce.JavascriptFunction;
import com.basho.riak.client.request.MapReduceBuilder;
import com.basho.riak.client.response.BucketResponse;
import com.basho.riak.client.response.FetchResponse;
import com.basho.riak.client.response.HttpResponse;
import com.basho.riak.client.response.MapReduceResponse;
import com.basho.riak.client.response.RiakIORuntimeException;
import com.basho.riak.client.response.RiakResponseRuntimeException;
import com.basho.riak.client.response.StoreResponse;

public final class RiakCache<T extends CachedItem> extends PersistentCache<T> {
    static public final Logger logger = Logger.getLogger(RiakCache.class);
    
    static private final String riakHost;
    
    static {
        Properties props = new Properties();
        
        try {
            InputStream is = DaseinSequencer.class.getResourceAsStream(DaseinSequencer.PROPERTIES);

            if( is != null ) {
                props.load(is);
            }
        }
        catch( Exception e ) {
            e.printStackTrace();
        }
        riakHost = props.getProperty("dasein.persist.riak.host");
    }
    
    public RiakCache(Class<T> cls, String pk) {
        this(cls, new Key(pk));
    }
    
    public RiakCache(Class<T> cls, Key ... keys) {
        super(cls, keys);
    }

    private void buildIndex(RiakClient client, String pk, Map<String,Object> state) throws PersistenceException {
        Key[] secondaryKeys = getSecondaryKeys();
        
        if( secondaryKeys != null && secondaryKeys.length > 0 ) {
            RiakLink link = new RiakLink(getTarget().getName(), pk, "index");
            
            for( Key k : secondaryKeys ) {
                String keyValue = getKeyValue(state, k);
                String bucketName = getBucketName(k);
                StoreResponse response;
                RiakObject index;
                
                FetchResponse fetch = client.fetch(bucketName, keyValue.toString());
                
                if( fetch.getStatusCode() == 404 || fetch.getObject() == null ) {
                    index = new RiakObject(client, bucketName, keyValue, new byte[0], "text/plain");
                    index.addLink(link);
                    response = client.store(index);
                }
                else {
                    boolean found = false;
                    
                    index = fetch.getObject();
                    for( RiakLink l : index.getLinks() ) {
                        if( l.getTag().equals(link.getTag()) && l.getKey().equals(link.getKey()) && l.getBucket().equals(link.getBucket()) ) {
                            found = true;
                            break;
                        }
                    }
                    if( found ) {
                        continue;
                    }
                    ArrayList<RiakLink> links = new ArrayList<RiakLink>();
                    
                    links.add(link);
                    index.setLinks(links);
                    response = client.store(index);
                }
                
                if( !response.isSuccess() ) {
                    logger.warn("Failed to build index on " + k + ": " + response.getStatusCode() + ": " + response.getBodyAsString());
                }
            }
        }        
    }
    
    public T create(Transaction xaction, Map<String,Object> state) throws PersistenceException {
        try {
            String key = getPrimaryKeyField();
            Object ob = state.get(key);
            
            if( ob == null ) {
                throw new PersistenceException("No value specified for key " + key);
            }
            
            String json = toJson(state);
            StoreResponse response = null;
            
            try {
                RiakClient client = getClient();
                RiakObject object;
                
                try {
                    object = new RiakObject(client, getTarget().getName(), ob.toString(), json.getBytes("utf-8"), "application/json");
                }
                catch( UnsupportedEncodingException e ) {
                    throw new PersistenceException(e);
                }
                response = client.store(object);
                if( response.isError() ) {
                    throw new PersistenceException(response.getStatusCode() + ": " + response.getBodyAsString());
                }
                buildIndex(client, ob.toString(), state);
            }
            finally {
                if( response != null ) {
                    response.close();
                }
            }
            T item = fromJson(json);
            
            getCache().cache(item);
            return item;
        }
        catch( RiakIORuntimeException e ) {
            throw new PersistenceException(e);
        }
        catch( RiakResponseRuntimeException e ) {
            throw new PersistenceException(e);
        }
    }
    
    private String escapeJson(String txt) {
        StringBuilder str = new StringBuilder();
        
        for( int i=0; i<txt.length(); i++ ) {
            char c = txt.charAt(i);
            
            switch( c ) {
                case '"': str.append("\\\""); break;
                case '\'': str.append("\\'"); break;
                default: str.append(c);
            }
        }
        return str.toString();
    }
    
    private String getMapReduce(SearchTerm ... terms) throws PersistenceException {
        if( terms != null && terms.length > 0 ) {
            StringBuilder tmp = new StringBuilder();
            
            for( SearchTerm term : terms ) {
                String operator;
            
                switch( term.getOperator() ) {
                    case EQUALS: operator = "=="; break;
                    case NOT_EQUAL: operator = "!="; break;
                    case GREATER_THAN: operator = ">"; break;
                    case LESS_THAN: operator = "<"; break;
                    case GREATER_THAN_OR_EQUAL_TO: operator = ">="; break;
                    case LESS_THAN_OR_EQUAL_TO: operator = "<="; break;
                    default: operator = "=="; break;
                }
                if( tmp.length() > 0 ) {
                    tmp.append(" && ");
                }
                tmp.append("item[\"");
                tmp.append(term.getColumn());
                tmp.append("\"] ");
                tmp.append(operator);
                tmp.append(toJson(term.getValue()));
            }
            
            return "if( value.values ) { var item = Riak.mapValuesJson(value)[0]; if( " + tmp.toString() + ") { return [item]; } else { return []; } } return [];";
        }
        else {
            return "if( value.values ) { var item = o = Riak.mapValuesJson(value)[0] return [item]; } return [];";
        }
    }
    
    @Override
    public Collection<T> find(SearchTerm[] terms, JiteratorFilter<T> filter, Boolean orderDesc, String ... orderFields) throws PersistenceException {
        Key key = matchKeys(terms);
        Set<String> keys = null;
        String bucket = null;
                
        if( key != null ) {
            if( key.equals(getPrimaryKey()) && !(terms[0].getValue() instanceof Range) ) {
                ArrayList<T> list = new ArrayList<T>();
                
                list.add(get(terms[0].getValue()));
                return list;
            }
            else {
                keys = getKeyValues(0, terms);
                if( !key.equals(getPrimaryKey()) ) {
                    bucket = getBucketName(key);
                }
            }
        }
        if( bucket == null ) {
            bucket = getTarget().getName();
        }
        final String fbucket = bucket;
        final Set<String> fkeys = keys;
        final SearchTerm[] fterms = terms;
        final JiteratorFilter<T> ffilter = filter;
        final boolean fdesc = (orderDesc == null ? false : orderDesc);
        final String[] ffields = orderFields;
        
        PopulatorThread<T> populator = new PopulatorThread<T>(new JiteratorPopulator<T>() {
            public void populate(Jiterator<T> it) throws PersistenceException {
                try {
                    boolean sorting = (ffields != null && ffields.length > 0);
                    RiakClient client = getClient();            
                    MapReduceBuilder builder;
        
                    if( fkeys == null ) {
                        String func = "function(value, meta, arg) { " + getMapReduce(fterms) + " }";
                        builder = client.mapReduceOverBucket(fbucket);                        
                        builder = builder.map(JavascriptFunction.anon(func), false);
                    }
                    else {
                        HashMap<String,Set<String>> inputs = new HashMap<String,Set<String>>();
                        
                        inputs.put(fbucket, fkeys);
                        builder = client.mapReduceOverObjects(inputs);
                        if( !fbucket.equals(getTarget().getName()) ) {
                            builder = builder.link(getTarget().getName(), false);
                        }
                    }
                    builder = builder.map(JavascriptFunction.named("Riak.filterNotFound"), !sorting);
                    if( sorting ) {
                        StringBuilder sort = new StringBuilder();
                        
                        sort.append("function(values, arg) { ");
                        sort.append("    return values.sort(function(a,b) { var x; ");
                        for( String field : ffields ) {
                            if( isNumber(field) ) {
                                if( fdesc ) {
                                    sort.append("x = b[\"");
                                    sort.append(field);
                                    sort.append("\"] - a[\"");
                                    sort.append(field);
                                    sort.append("\"];");
                                }
                                else {
                                    sort.append("x = a[\"");
                                    sort.append(field);
                                    sort.append("\"] - b[\"");
                                    sort.append(field);
                                    sort.append("\"];");                                   
                                }
                            }
                            else {
                                if( fdesc ) {
                                    sort.append("if( b[\"");
                                    sort.append(field);
                                    sort.append("\"] > a[\"");
                                    sort.append(field);
                                    sort.append("\"] ) { x = 1; } else if( a[\"");
                                    sort.append(field);
                                    sort.append("\"] > b[\"");
                                    sort.append(field);
                                    sort.append("\"] ) { x = -1; } else { x = 0; }");
                                }
                                else {
                                    sort.append("if( a[\"");
                                    sort.append(field);
                                    sort.append("\"] > b[\"");
                                    sort.append(field);
                                    sort.append("\"] ) { x = 1; } else if( b[\"");
                                    sort.append(field);
                                    sort.append("\"] > a[\"");
                                    sort.append(field);
                                    sort.append("\"] ) { x = -1; } else { x = 0; }");                                    
                                }
                            }
                            sort.append(" if( x != 0 ) { return x; } ");
                        }
                        sort.append("    return 0; });");
                        sort.append("}");
                        System.out.println(sort.toString());
                        builder.reduce(JavascriptFunction.anon(sort.toString()), true);
                    }
                    MapReduceResponse response = builder.submit();
                    
                    if( response.isSuccess() ) {
                        JSONArray results = response.getResults();
                        int len = results.length();

                        for( int i=0; i<len; i++ ) {
                            T item = fromJson(results.getString(i));

                            if( ffilter == null || ffilter.filter(item) ) {
                                it.push(item);
                            }
                        }
                    }
                    else {
                        throw new PersistenceException(response.getStatusCode() + ": " + response.getBodyAsString());
                    }
                }
                catch( RiakIORuntimeException e ) {
                    throw new PersistenceException(e);
                }
                catch( RiakResponseRuntimeException e ) {
                    throw new PersistenceException(e);
                }
                catch( JSONException e ) {
                    throw new PersistenceException(e);
                } 
                catch( Throwable t ) {
                    t.printStackTrace();
                    throw new PersistenceException(t.getMessage());
                }
            }
        });
        populator.populate();
        return populator.getResult();

    }
    
    @SuppressWarnings("unchecked")
    private T fromJson(String json) throws PersistenceException {
        try {
            JSONObject j = JSONObject.fromObject(json);
            Class<T> targetClass = getTarget();
            Class<? extends Object> t = targetClass;
            T object;
            
            try {
                object = targetClass.newInstance();
            }
            catch( Exception e ) {
                throw new PersistenceException(e);
            }
            while( !t.getName().equals(Object.class.getName()) ) {
                for( Field field : t.getDeclaredFields() ) {
                    int modifiers = field.getModifiers();
                    
                    if( !Modifier.isStatic(modifiers) && !Modifier.isTransient(modifiers) ) {
                        Class<?> type = field.getType();
                        
                        if( !j.isNullObject() ) {
                            field.setAccessible(true);
                            try {
                                if( type.equals(String.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, null);
                                    }
                                    else {
                                        field.set(object, j.getString(field.getName()));
                                    }
                                }
                                if( Enum.class.isAssignableFrom(type) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, null);
                                    }
                                    else {
                                        String value = j.getString(field.getName());
                                        
                                        if( value == null ) {
                                            field.set(object, null);
                                        }
                                        else {
                                            field.set(object, Enum.valueOf((Class<? extends Enum>)type, value));
                                        }
                                    }
                                }
                                else if( type.equals(Boolean.class) || type.equals(boolean.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, false);
                                    }
                                    else {
                                        field.set(object, j.getBoolean(field.getName()));
                                    }
                                }
                                else if( type.equals(Integer.class) || type.equals(int.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, type.equals(Integer.class) ? null : 0);
                                    }
                                    else {
                                        field.set(object, j.getInt(field.getName()));
                                    }
                                }
                                else if( type.equals(Short.class) || type.equals(short.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, type.equals(Short.class) ? null : 0);
                                    }
                                    else {
                                        field.set(object, (short)j.getInt(field.getName()));
                                    }
                                }
                                else if( type.equals(Long.class) || type.equals(long.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, type.equals(Long.class) ? null : 0L);
                                    }
                                    else {
                                        field.set(object, j.getLong(field.getName()));
                                    }
                                }
                                else if( type.equals(Float.class) || type.equals(float.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, type.equals(Float.class) ? null : 0.0f);
                                    }
                                    else {
                                        field.set(object, (float)j.getDouble(field.getName()));
                                    }
                                }
                                else if( type.equals(Double.class) || type.equals(double.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, type.equals(Double.class) ? null : 0.0);
                                    }
                                    else {
                                        field.set(object, j.getDouble(field.getName()));
                                    }
                                } 
                                else if( Number.class.isAssignableFrom(type) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, 0.0);
                                    }
                                    else {
                                        Object ob = j.get(field.getName());
                                    
                                        if( ob instanceof Long ) {
                                            field.set(object, j.getLong(field.getName()));
                                        }
                                        else if( ob instanceof Integer ) {
                                            field.set(object, j.getInt(field.getName()));
                                        }
                                        else if( ob instanceof Double ) {
                                            field.set(object, j.getDouble(field.getName()));
                                        }
                                    }
                                }      
                                else if( type.equals(Locale.class )) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, null);
                                    }
                                    else {
                                        String[] parts = j.getString(field.getName()).split("_");
                                        Locale loc;
                                        
                                        if( parts != null && parts.length > 1 ) {
                                            loc = new Locale(parts[0], parts[1]);
                                        }
                                        else {
                                            loc = new Locale(parts[0]);
                                        }      
                                        field.set(object, loc);
                                    }
                                }
                                else if( type.equals(UUID.class) ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, null);
                                    }
                                    else {
                                        field.set(object, UUID.fromString(j.getString(field.getName())));
                                    }
                                }
                                else if( type.getName().startsWith("java.") ) {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, null);
                                    }
                                    else {
                                        field.set(object, j.get(field.getName()));
                                    }
                                }
                                else {
                                    if( !j.has(field.getName()) ) {
                                        field.set(object, null);
                                    }
                                    else {
                                        String value = j.getString(field.getName());
                                        
                                        try {
                                            Method m = type.getDeclaredMethod("valueOf", String.class);
                                            
                                            if( m != null ) {
                                                field.set(object, m.invoke(null, value));
                                            }
                                            else {
                                                throw new PersistenceException("No valueOf() field in " + type + " for mapping " + field.getName());
                                            }
                                        }
                                        catch( Exception e ) {
                                            throw new PersistenceException("Error mapping field in " + type + " for " + field.getName() + ": " + e.getMessage());
                                        }
                                    }
                                }
                            }
                            catch( Exception e ) {
                                throw new PersistenceException(e);
                            }
                        }
                    }
                }
                t = t.getSuperclass();
            }
            T item = getCache().find(getPrimaryKeyField(), getKeyValue(object));
            
            if( item != null ) {
                return item;
            }
            getCache().cache(object);
            return object;
        }
        catch( RuntimeException e ) {
            e.printStackTrace();
            throw new PersistenceException(e);
        }
        catch( Error e ) {
            e.printStackTrace();
            throw new PersistenceException(e.getMessage());
        }
    }
    
    @Override
    public T get(Object valueOfPrimaryKey) throws PersistenceException {
        try {
            T item = getCache().find(getPrimaryKeyField(), valueOfPrimaryKey.toString());
            
            if( item == null ) {
                RiakClient client = getClient();
                FetchResponse response = null;
                
                try {
                    response = client.fetch(getCache().getTarget().getName(), valueOfPrimaryKey.toString());
                    if( response.isSuccess() ) {
                        RiakObject object = response.getObject();
                        
                        if( object == null ) {
                            return null;
                        }
                        item = fromJson(object.getValue());
                        getCache().cache(item);
                    }
                    else {
                        throw new PersistenceException(response.getStatusCode() + ": " + response.getBodyAsString());
                    }
                }
                finally {
                    if( response != null ) {
                        response.close();
                    }
                }
            }
            return item;
        }
        catch( RiakIORuntimeException e ) {
            throw new PersistenceException(e);
        }
        catch( RiakResponseRuntimeException e ) {
            throw new PersistenceException(e);
        }
    }    

    private String getBucketName(Key key) {
        String bucket = null;
        
        for( String field : key.getFields() ) {
            if( bucket == null ) {
                bucket = getCache().getTarget().getName() + "[" + field;
            }
            else {
                bucket = bucket + "," + field;
            }
        }
        bucket = bucket + "]"; 
        return bucket;
    }
    
    private RiakClient getClient() throws PersistenceException {
        return new RiakClient(riakHost);
    }
    
    private boolean isNumber(String field) throws PersistenceException {
        Class<? extends Object> t = getTarget();
        
        while( !t.equals(Object.class) ) {
            Field f;

            try {
                f = t.getDeclaredField(field);
            }
            catch( SecurityException e ) {
                throw new PersistenceException(e);
            }
            catch( NoSuchFieldException e ) {
                continue;
            }
            if( f != null ) {
                return Number.class.isAssignableFrom(f.getType());
            }
        }
        return false;
    }
    
    @Override
    public Collection<T> list() throws PersistenceException {
        PopulatorThread<T> populator;
        
        populator = new PopulatorThread<T>(new JiteratorPopulator<T>() {
            public void populate(Jiterator<T> iterator) throws PersistenceException {
                try {
                    BucketResponse response = null;
    
                    try {
                        response = getClient().streamBucket(getTarget().getName());
                        if( response.isSuccess() ) {
                            for( String key : response.getBucketInfo().getKeys() ) {
                                T object = get(key);
                                
                                if( object != null ) {
                                    iterator.push(object);
                                }
                            }
                        }
                        else {
                            throw new PersistenceException(response.getStatusCode() + ": " + response.getBodyAsString());
                        }
                    } 
                    finally {
                        if( response != null ) {
                            response.close();
                        }
                    }
                }
                catch( RiakIORuntimeException e ) {
                    throw new PersistenceException(e);
                }
                catch( RiakResponseRuntimeException e ) {
                    throw new PersistenceException(e);
                }
            }
        });
        populator.populate();
        return populator.getResult();
    }
    
    @Override
    public void remove(Transaction xaction, T item) throws PersistenceException {
        try {
            HttpResponse response = null;
            
            try {
                RiakClient client = getClient();
                
                response = client.delete(getTarget().getName(), getKeyValue(item));
                if( response.isError() ) {
                    throw new PersistenceException(response.getStatusCode() + ": " + response.getBodyAsString());
                }
                removeIndex(client, item);
            }
            finally {
                if( response != null ) {
                    response.close();
                }
                getCache().release(item);
            }
        }
        catch( RiakIORuntimeException e ) {
            throw new PersistenceException(e);
        }
        catch( RiakResponseRuntimeException e ) {
            throw new PersistenceException(e);
        }    
    }
    
    private void removeIndex(RiakClient client, T item) throws PersistenceException {
        Key[] secondaryKeys = getSecondaryKeys();
        
        if( secondaryKeys != null && secondaryKeys.length > 0 ) {
            for( Key k : secondaryKeys ) {
                String keyValue = getKeyValue(item, k);
                String bucketName = getBucketName(k);
                HttpResponse response;
                
                FetchResponse fetch = client.fetch(bucketName, keyValue.toString());
                
                if( fetch.getStatusCode() == 404 || fetch.getObject() == null ) {
                    continue;
                }
                else {
                    response = client.delete(bucketName, keyValue);
                }                
                if( !response.isSuccess() ) {
                    logger.warn("Failed to remove index on " + k + ": " + response.getStatusCode() + ": " + response.getBodyAsString());
                }
            }
        }        
    }
    
    @SuppressWarnings("unchecked")
    private String toJson(Object value) {
        if( value instanceof String ) {
            return "\"" + escapeJson((String)value) + "\"";
        }
        else if( Number.class.isAssignableFrom(value.getClass()) ) {
            return value.toString();
        }
        else if( value instanceof Enum ) {
            return "\"" + ((Enum)value).name() + "\"";
        }
        else if( value instanceof Locale ) {
            Locale loc = (Locale)value;
            String str = loc.getLanguage();
            
            if( loc.getCountry() != null ) {
                str = str + "_" + loc.getCountry();
            }
            return "\"" + str + "\"";
        }
        else if( value instanceof UUID ) {
            return "\"" + ((UUID)value).toString() + "\"";
        }
        else {
            return "\"" + escapeJson(value.toString()) + "\"";
        }
    }
    
    private String toJson(Map<String,Object> state) {
        StringBuilder json = new StringBuilder();
        Class<? extends Object> t = getTarget();
        boolean first = true;
        
        json.append("{");
        while( !t.equals(Object.class) ) {
            for( Field field : t.getDeclaredFields()) {
                field.setAccessible(true);
                
                int modifiers = field.getModifiers();
                
                if( Modifier.isStatic(modifiers) || Modifier.isTransient(modifiers) ) {
                    continue;
                }
                Object value = state.get(field.getName());
                
                if( value != null ) {
                    if( !first ) {
                        json.append(", ");
                    }
                    first = false;
                    json.append("\"");
                    json.append(field.getName());
                    json.append("\" : ");
                    json.append(toJson(value));
                }
            }
            t = t.getSuperclass();
        }
        json.append("}");
        return json.toString();
    }
    
    public String toString() {
        return getCache().toString();
    }
 
    @Override
    public void update(Transaction xaction, T item, Map<String,Object> state) throws PersistenceException {
        String pk = getKeyValue(state, getPrimaryKey());
        RiakClient client = getClient();
        
        try {
            FetchResponse response = null;
            
            try {
                response = client.fetch(getTarget().getName(), pk);
                if( response.isSuccess() ) {
                    RiakObject object = response.getObject();
                
                    object.setValue(toJson(state));
                    client.store(object);
                    buildIndex(client, pk, state);
                }
                else if( response.isError() ) {
                    throw new PersistenceException(response.getStatusCode() + ": " + response.getBodyAsString());
                }
            }
            finally {
                if( response != null ) {
                    response.close();
                }
            }
        }
        catch( RiakIORuntimeException e ) {
            throw new PersistenceException(e);
        }
        catch( RiakResponseRuntimeException e ) {
            throw new PersistenceException(e);
        }
    }    
}
