/* $Id: Loader.java,v 1.16 2009/02/21 19:44:38 greese Exp $ */
/* Copyright (c) 2006 Valtira Corporation, All Rights Reserved */
package org.dasein.persist.jdbc;

import java.lang.reflect.Method;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.UUID;

import org.apache.log4j.Logger;
import org.dasein.persist.PersistenceException;
import org.dasein.persist.Transaction;
import org.dasein.persist.l10n.LocalizationGroup;

public class Loader extends AutomatedSql {
    static public final Logger logger = Logger.getLogger(Loader.class);
    
    private boolean                  descending = false;
    private ArrayList<String>        order      = new ArrayList<String>();
    private String                   sql        = null;
    
    public Loader() {
        super();
    }
    
    public synchronized String getStatement() throws SQLException {
        logger.debug("enter - getStatement()");
        try {
            if( sql == null ) {
                StringBuffer str = new StringBuffer();
                Iterator<String> it = getColumns().iterator();
    
                str.append("SELECT ");
                while( it.hasNext() ) {
                    String col = it.next();
                    
                    str.append(getIdentifier(getTableName(), getSqlName(col)));
                    if( it.hasNext() ) {
                        str.append(", ");
                    }
                }
                str.append(" FROM ");
                str.append(getIdentifier(getTableName()));
                if( !getCriteria().isEmpty() ) {
                    Iterator<Criterion> criteria;
                    
                    str.append(" WHERE ");
                    criteria = getCriteria().iterator();
                    while( criteria.hasNext() ) {
                        Criterion criterion = criteria.next();
                        
                        str.append(getIdentifier(getTableName(), getSqlName(criterion.column)));
                        str.append(" ");
                        str.append(criterion.operator.toString());
                        str.append(" ?");
                        if( criteria.hasNext() ) {
                            str.append(" ");
                            str.append(getJoin().toString());
                            str.append(" ");
                        }
                    }
                }
                if( order.size() > 0 ) {
                    str.append(" ORDER BY ");
                    it = order.iterator();
                    while( it.hasNext() ) {
                        String col = it.next();
                        
                        str.append(getIdentifier(getTableName(), getSqlName(col)));
                        if( it.hasNext() ) {
                            str.append(", ");
                        }
                    }
                    if( descending ) {
                        str.append(" DESC ");
                    }
                }
                sql = str.toString();
            }
            return sql;
        }
        finally {
            logger.debug("exit - getStatement()");
        }
    }

    protected void setOrder(boolean desc, String... cols) {
        descending = desc;
        for( String col : cols ) {
            order.add(col);
        }
    }
    
    public void prepare(Map<String,Object> params) throws SQLException {
        int i = 1;
        
        for( Criterion criterion : getCriteria() ) {
            prepare(criterion.column, i++, params.get(criterion.column));
        }
    }
    
    public Map<String,Object> run(Transaction xaction, Map<String,Object> params) throws SQLException, PersistenceException {
        ArrayList<Map<String,Object>> list = new ArrayList<Map<String,Object>>();
        HashMap<String,Object> map = new HashMap<String,Object>();
        int count = getColumns().size();
        long startTimestamp = System.currentTimeMillis();
        
        map.put("listing", list);
        prepare(params);
        results = statement.executeQuery();
        while( results.next() ) {
            HashMap<String,Object> state = new HashMap<String,Object>();
            
            for( int i=1; i<=count; i++) {
                Object ob = getValue(getColumns().get(i-1), i, results);
                
                state.put(getColumns().get(i-1), ob);
            }
            list.add(state);
        }
        long endTimestamp = System.currentTimeMillis();
        if( (endTimestamp - startTimestamp) > (2000L) ) {
            logger.warn("SLOW QUERY: " + sql);
        }
        if( isTranslating() ) {
            for( Map<String,Object> item : list ) {
                Object key = item.get((String)params.get("--key--"));
                
                item.putAll(loadStringTranslations(xaction, getTarget(), key.toString()));
            }            
        }
        return map;
    }
    
    @SuppressWarnings("unchecked")
    public Object getValue(String col, int i, ResultSet rs) throws SQLException {
        Class<?> type = getTypes().get(col);
        Object ob;
        
        if( type.equals(String.class) ) {
            String str = rs.getString(i);
            
            if( rs.wasNull() || str == null ) {
                ob = null;
            }
            else {
                ob = str.trim();
            }
        }
        else if( type.equals(Boolean.class) ) {
            String str = rs.getString(i);
            
            if( rs.wasNull() || str == null ) {
                ob = false;
            }
            else {
                ob = str.equalsIgnoreCase("Y");
            }
        }
        else if( type.equals(Locale.class) ) {
            String str = rs.getString(i);
            
            if( rs.wasNull() || str == null ) {
                ob = null;
            }
            else {
                String[] parts = str.split("_");
                
                if( parts != null && parts.length > 1 ) {
                    ob = new Locale(parts[0], parts[1]);
                }
                else {
                    ob = new Locale(parts[0]);
                }
            }
        }
        else if( type.equals(LocalizationGroup.class) ) {
            String str = rs.getString(i);
            
            if( rs.wasNull() || str == null ) {
                ob = null;
            }
            else {
                ob = LocalizationGroup.valueOf(str);
            }
        }
        else if( Number.class.isAssignableFrom(type) ) {
            if( type.getName().equals(Long.class.getName()) ) {
                long l = rs.getLong(i);
                
                if( rs.wasNull() ) {
                    ob = null;
                }
                else {
                    ob = l;
                }
            }
            else if( type.getName().equals(Integer.class.getName()) || type.getName().equals(Short.class.getName()) ) {
                int x = rs.getInt(i);
                
                if( rs.wasNull() ) {
                    ob = null;
                }
                else {
                    ob = x;
                }
            }
            else if( type.getName().equals(Double.class.getName()) ) {
                double x = rs.getDouble(i);

                if( rs.wasNull() ) {
                    ob = null;
                }
                else {
                    ob = x;
                }
            }
            else if( type.getName().equals(Float.class.getName()) ) {
                float f = rs.getFloat(i);
                
                if( rs.wasNull() ) {
                    ob = null;
                }
                else {
                    ob = f;
                }
            }
            else {
                ob = rs.getBigDecimal(i);
                if( rs.wasNull() ) {
                    ob = null;
                }
            }
        }
        else if( Enum.class.isAssignableFrom(type) ) {
            String str = rs.getString(i);
            
            if( str == null || rs.wasNull() ) {
                ob = null;
            }
            else {
                ob = Enum.valueOf((Class<? extends Enum>)type, str);
            }
        }
        else if( type.equals(UUID.class) ) {
            String str = rs.getString(i);
            
            if( rs.wasNull() || str == null ) {
                ob = null;
            }
            else {
                ob = UUID.fromString(str);
            }
        }
        else if( type.getName().startsWith("java.") ){
            ob = rs.getObject(i);
            if( rs.wasNull() ) {
                ob = null;
            }
        }
        else {
            String str = rs.getString(i);

            if( str == null || rs.wasNull() ) {
                ob = null;
            }
            else {
                try {
                    Method m = type.getDeclaredMethod("valueOf", String.class);
                    
                    ob = m.invoke(null, str);
                }
                catch( Exception e ) {
                    throw new SQLException("I have no idea how to map to " + type.getName());
                }
            }
        }
        return ob;
    }
}
